#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ******************************************************************************
#
#  Project:  GDAL
#  Purpose:  Make fuzzer friendly archive (only works in DEBUG mode)
#  Author:   Even Rouault, <even dot rouault at spatialys dot com>
#
# ******************************************************************************
#  Copyright (c) 2016 Even Rouault, <even dot rouault at spatialys dot com>
#
# SPDX-License-Identifier: MIT
# ******************************************************************************

import os
import sys


def Usage():
    print(
        f"Usage: {sys.argv[0]} -- This is a sample. Read source to know how to use. --"
    )
    return 2


def main(argv=sys.argv):
    if len(sys.argv) < 2:
        return Usage()
    fout = open(argv[1], "wb")
    fout.write("FUZZER_FRIENDLY_ARCHIVE\n".encode("ascii"))
    for filename in argv[2:]:
        fout.write(("***NEWFILE***:%s\n" % os.path.basename(filename)).encode("ascii"))
        fout.write(open(filename, "rb").read())
    fout.close()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# ******************************************************************************
#
#  Project:  GDAL
#  Purpose:  Example computing the magnitude and phase from a complex image.
#  Author:   Frank Warmerdam, warmerdam@pobox.com
#
# ******************************************************************************
#  Copyright (c) 2008, Frank Warmerdam <warmerdam@pobox.com>
#
# SPDX-License-Identifier: MIT
# ******************************************************************************

import sys

import numpy as np

from osgeo import gdal, gdal_array


def Usage():
    print(
        f"Usage: {sys.argv[0]} -- This is a sample. Read source to know how to use. --"
    )
    return 2


def doit(src_filename, dst_magnitude, dst_phase):
    src_ds = gdal.Open(src_filename)
    xsize = src_ds.RasterXSize
    ysize = src_ds.RasterYSize
    print("{} x {}".format(xsize, ysize))

    src_image = src_ds.GetRasterBand(1).ReadAsArray()
    mag_image = pow(
        np.real(src_image) * np.real(src_image)
        + np.imag(src_image) * np.imag(src_image),
        0.5,
    )
    gdal_array.SaveArray(mag_image, dst_magnitude)

    phase_image = np.angle(src_image)
    gdal_array.SaveArray(phase_image, dst_phase)
    return 0


def main(argv=sys.argv):
    if len(sys.argv) <= 4:
        return Usage()
    # src_filename = 'complex.tif'
    # dst_magnitude = 'magnitude.tif'
    # dst_phase = 'phase.tif'
    if len(argv) > 1:
        src_filename = argv[1]
    if len(argv) > 2:
        dst_magnitude = argv[2]
    if len(argv) > 3:
        dst_phase = argv[3]
    return doit(src_filename, dst_magnitude, dst_phase)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ******************************************************************************
#
#  Project:  GDAL Python Interface
#  Purpose:  Application for executing OGR layer algebra operations
#  Author:   Even Rouault, even dot rouault at spatialys.com
#
# ******************************************************************************
#  Copyright (c) 2013, Even Rouault <even dot rouault at spatialys.com>
#
# SPDX-License-Identifier: MIT
# ******************************************************************************

import os
import sys

from osgeo import gdal, ogr, osr
from osgeo_utils.auxiliary.util import enable_gdal_exceptions

###############################################################################


def Usage(isError):
    print(
        """
Usage: ogr_layer_algebra [--help] [--help-general]
                            Union|Intersection|SymDifference|Identity|Update|Clip|Erase
                            -input_ds <path> [-input_lyr <name>]
                            -method_ds <path> [-method_lyr <name>]
                            -output_ds <path> [-output_lyr name] [-overwrite]
                            [-opt <NAME>=<VALUE>]...
                            [-f <format_name>] [-dsco <NAME>=<VALUE>]... [-lco <NAME>=<VALUE>]...
                            [-input_fields {NONE|ALL|<fld1>,<fl2>,...<fldN>}] [-method_fields {NONE|ALL|<fld1>,<fl2>,...<fldN>}]
                            [-nlt <geom_type>] [-a_srs <srs_def>]""",
        file=sys.stderr if isError else sys.stdout,
    )
    return 2 if isError else 0


###############################################################################


def EQUAL(a, b):
    return a.lower() == b.lower()


###############################################################################


def CreateLayer(
    output_ds,
    output_lyr_name,
    srs,
    geom_type,
    lco,
    input_lyr,
    input_fields,
    method_lyr,
    method_fields,
    opt,
):

    output_lyr = output_ds.CreateLayer(output_lyr_name, srs, geom_type, lco)
    if output_lyr is None:
        print('Cannot create layer "%s"' % output_lyr_name, file=sys.stderr)
        return None

    if input_fields == "ALL" and method_fields == "ALL":
        return output_lyr

    input_prefix = ""
    method_prefix = ""
    for val in opt:
        if val.lower().find("input_prefix=") == 0:
            input_prefix = val[len("input_prefix=") :]
        elif val.lower().find("method_prefix=") == 0:
            method_prefix = val[len("method_prefix=") :]

    if input_fields == "ALL":
        layer_defn = input_lyr.GetLayerDefn()
        for idx in range(layer_defn.GetFieldCount()):
            fld_defn = layer_defn.GetFieldDefn(idx)
            fld_defn = ogr.FieldDefn(
                input_prefix + fld_defn.GetName(), fld_defn.GetType()
            )
            if output_lyr.CreateField(fld_defn) != 0:
                print(
                    'Cannot create field "%s" in layer "%s"'
                    % (fld_defn.GetName(), output_lyr.GetName()),
                    file=sys.stderr,
                )

    elif input_fields != "NONE":
        layer_defn = input_lyr.GetLayerDefn()
        for fld in input_fields:
            idx = layer_defn.GetFieldIndex(fld)
            if idx < 0:
                print(
                    'Cannot find field "%s" in layer "%s"'
                    % (fld, layer_defn.GetName()),
                    file=sys.stderr,
                )
                continue
            fld_defn = layer_defn.GetFieldDefn(idx)
            fld_defn = ogr.FieldDefn(
                input_prefix + fld_defn.GetName(), fld_defn.GetType()
            )
            if output_lyr.CreateField(fld_defn) != 0:
                print(
                    'Cannot create field "%s" in layer "%s"'
                    % (fld, output_lyr.GetName()),
                    file=sys.stderr,
                )

    if method_fields == "ALL":
        layer_defn = method_lyr.GetLayerDefn()
        for idx in range(layer_defn.GetFieldCount()):
            fld_defn = layer_defn.GetFieldDefn(idx)
            fld_defn = ogr.FieldDefn(
                method_prefix + fld_defn.GetName(), fld_defn.GetType()
            )
            if output_lyr.CreateField(fld_defn) != 0:
                print(
                    'Cannot create field "%s" in layer "%s"'
                    % (fld_defn.GetName(), output_lyr.GetName()),
                    file=sys.stderr,
                )

    elif method_fields != "NONE":
        layer_defn = method_lyr.GetLayerDefn()
        for fld in method_fields:
            idx = layer_defn.GetFieldIndex(fld)
            if idx < 0:
                print(
                    'Cannot find field "%s" in layer "%s"'
                    % (fld, layer_defn.GetName()),
                    file=sys.stderr,
                )
                continue
            fld_defn = layer_defn.GetFieldDefn(idx)
            fld_defn = ogr.FieldDefn(
                method_prefix + fld_defn.GetName(), fld_defn.GetType()
            )
            if output_lyr.CreateField(fld_defn) != 0:
                print(
                    'Cannot create field "%s" in layer "%s"'
                    % (fld, output_lyr.GetName()),
                    file=sys.stderr,
                )

    return output_lyr


###############################################################################


@enable_gdal_exceptions
def main(argv=sys.argv):

    driver_name = "ESRI Shapefile"
    quiet = False
    input_ds_name = None
    input_lyr_name = None
    method_ds_name = None
    method_lyr_name = None
    output_ds_name = None
    output_lyr_name = None
    op_str = None
    dsco = []
    lco = []
    opt = []
    overwrite = False
    input_fields = "ALL"
    method_fields = None
    geom_type = ogr.wkbUnknown
    srs_name = None
    srs = None

    argv = ogr.GeneralCmdLineProcessor(argv)
    if argv is None:
        return 0

    # Parse command line arguments.
    i = 1
    while i < len(argv):
        arg = argv[i]

        if arg == "--help":
            return Usage(isError=False)

        elif arg == "-f" and i + 1 < len(argv):
            i = i + 1
            driver_name = argv[i]

        elif arg == "-input_ds" and i + 1 < len(argv):
            i = i + 1
            input_ds_name = argv[i]

        elif arg == "-input_lyr" and i + 1 < len(argv):
            i = i + 1
            input_lyr_name = argv[i]

        elif arg == "-method_ds" and i + 1 < len(argv):
            i = i + 1
            method_ds_name = argv[i]

        elif arg == "-method_lyr" and i + 1 < len(argv):
            i = i + 1
            method_lyr_name = argv[i]

        elif arg == "-output_ds" and i + 1 < len(argv):
            i = i + 1
            output_ds_name = argv[i]

        elif arg == "-output_lyr" and i + 1 < len(argv):
            i = i + 1
            output_lyr_name = argv[i]

        elif arg == "-input_fields" and i + 1 < len(argv):
            i = i + 1
            if EQUAL(argv[i], "NONE"):
                input_fields = "NONE"
            elif EQUAL(argv[i], "ALL"):
                input_fields = "ALL"
            else:
                input_fields = argv[i].split(",")

        elif arg == "-method_fields" and i + 1 < len(argv):
            i = i + 1
            if EQUAL(argv[i], "NONE"):
                method_fields = "NONE"
            elif EQUAL(argv[i], "ALL"):
                method_fields = "ALL"
            else:
                method_fields = argv[i].split(",")

        elif arg == "-dsco" and i + 1 < len(argv):
            i = i + 1
            dsco.append(argv[i])

        elif arg == "-lco" and i + 1 < len(argv):
            i = i + 1
            lco.append(argv[i])

        elif arg == "-opt" and i + 1 < len(argv):
            i = i + 1
            opt.append(argv[i])

        elif arg == "-nlt" and i + 1 < len(argv):
            i = i + 1
            val = argv[i]

            if EQUAL(val, "NONE"):
                geom_type = ogr.wkbNone
            elif EQUAL(val, "GEOMETRY"):
                geom_type = ogr.wkbUnknown
            elif EQUAL(val, "POINT"):
                geom_type = ogr.wkbPoint
            elif EQUAL(val, "LINESTRING"):
                geom_type = ogr.wkbLineString
            elif EQUAL(val, "POLYGON"):
                geom_type = ogr.wkbPolygon
            elif EQUAL(val, "GEOMETRYCOLLECTION"):
                geom_type = ogr.wkbGeometryCollection
            elif EQUAL(val, "MULTIPOINT"):
                geom_type = ogr.wkbMultiPoint
            elif EQUAL(val, "MULTILINESTRING"):
                geom_type = ogr.wkbMultiLineString
            elif EQUAL(val, "MULTIPOLYGON"):
                geom_type = ogr.wkbMultiPolygon
            elif EQUAL(val, "GEOMETRY25D"):
                geom_type = ogr.wkbUnknown | ogr.wkb25DBit
            elif EQUAL(val, "POINT25D"):
                geom_type = ogr.wkbPoint25D
            elif EQUAL(val, "LINESTRING25D"):
                geom_type = ogr.wkbLineString25D
            elif EQUAL(val, "POLYGON25D"):
                geom_type = ogr.wkbPolygon25D
            elif EQUAL(val, "GEOMETRYCOLLECTION25D"):
                geom_type = ogr.wkbGeometryCollection25D
            elif EQUAL(val, "MULTIPOINT25D"):
                geom_type = ogr.wkbMultiPoint25D
            elif EQUAL(val, "MULTILINESTRING25D"):
                geom_type = ogr.wkbMultiLineString25D
            elif EQUAL(val, "MULTIPOLYGON25D"):
                geom_type = ogr.wkbMultiPolygon25D
            else:
                print("-nlt %s: type not recognised." % val, file=sys.stderr)
                return 1

        elif arg == "-a_srs" and i + 1 < len(argv):
            i = i + 1
            srs_name = argv[i]

        elif EQUAL(arg, "Union"):
            op_str = "Union"

        elif EQUAL(arg, "Intersection"):
            op_str = "Intersection"

        elif EQUAL(arg, "SymDifference"):
            op_str = "SymDifference"

        elif EQUAL(arg, "Identity"):
            op_str = "Identity"

        elif EQUAL(arg, "Update"):
            op_str = "Update"

        elif EQUAL(arg, "Clip"):
            op_str = "Clip"

        elif EQUAL(arg, "Erase"):
            op_str = "Erase"

        elif arg == "-overwrite":
            overwrite = True

        elif arg == "-q" or arg == "-quiet":
            quiet = True

        else:
            return Usage(isError=True)

        i = i + 1

    if (
        input_ds_name is None
        or method_ds_name is None
        or output_ds_name is None
        or op_str is None
    ):
        return Usage(isError=True)

    if method_fields is None:
        if op_str in ("Update", "Clip", "Erase"):
            method_fields = "NONE"
        else:
            method_fields = "ALL"

    if input_fields == "NONE" and method_fields == "NONE":
        print(
            "Warning: -input_fields NONE and -method_fields NONE results in all fields being added",
            file=sys.stderr,
        )

    # Input layer
    input_ds = ogr.Open(input_ds_name)
    if input_ds is None:
        print("Cannot open input dataset : %s" % input_ds_name, file=sys.stderr)
        return 1

    if input_lyr_name is None:
        cnt = input_ds.GetLayerCount()
        if cnt != 1:
            print(
                "Input datasource has more than just a single layer, so you should specify which layer with -input_lyr",
                file=sys.stderr,
            )
            return 1
        input_lyr = input_ds.GetLayer(0)
    else:
        input_lyr = input_ds.GetLayerByName(input_lyr_name)
    if input_lyr is None:
        print('Cannot find input layer "%s"' % input_lyr_name, file=sys.stderr)
        return 1

    # Method layer
    method_ds = ogr.Open(method_ds_name)
    if method_ds is None:
        print("Cannot open method dataset : %s" % method_ds_name, file=sys.stderr)
        return 1

    if method_lyr_name is None:
        cnt = method_ds.GetLayerCount()
        if cnt != 1:
            print(
                "Method datasource has not a single layer, so you should specify its name with -method_lyr",
                file=sys.stderr,
            )
            return 1
        method_lyr = method_ds.GetLayer(0)
    else:
        method_lyr = method_ds.GetLayerByName(method_lyr_name)
    if method_lyr is None:
        print('Cannot find method layer "%s"' % method_lyr_name, file=sys.stderr)
        return 1

    # SRS
    if srs_name is not None:
        if not EQUAL(srs_name, "NULL") and not EQUAL(srs_name, "NONE"):
            srs = osr.SpatialReference()
            if srs.SetFromUserInput(srs_name) != 0:
                print(
                    "Failed to process SRS definition: %s" % srs_name, file=sys.stderr
                )
                return 1
    else:
        srs = input_lyr.GetSpatialRef()
        srs2 = method_lyr.GetSpatialRef()
        if srs is None and srs2 is not None:
            print(
                "Warning: input layer has no SRS defined, but method layer has one.",
                file=sys.stderr,
            )
        elif srs is not None and srs2 is None:
            print(
                "Warning: input layer has a SRS defined, but method layer has not one.",
                file=sys.stderr,
            )
        elif srs is not None and srs2 is not None and srs.IsSame(srs2) != 1:
            print(
                "Warning: input and method layers have SRS defined, but they are not identical. No on-the-fly reprojection will be done.",
                file=sys.stderr,
            )

    # Result layer
    with ogr.ExceptionMgr(useExceptions=False), gdal.quiet_errors():
        output_ds = ogr.Open(output_ds_name, update=1)
    if output_ds is None:
        with ogr.ExceptionMgr(useExceptions=False), gdal.quiet_errors():
            output_ds = ogr.Open(output_ds_name)
        if output_ds is not None:
            print(
                'Output datasource "%s" exists, but cannot be opened in update mode'
                % output_ds_name,
                file=sys.stderr,
            )
            return 1

        drv = ogr.GetDriverByName(driver_name)
        if drv is None:
            print("Cannot find driver %s" % driver_name, file=sys.stderr)
            return 1

        output_ds = drv.CreateDataSource(output_ds_name, options=dsco)
        if output_ds is None:
            print('Cannot create datasource "%s"' % output_ds_name, file=sys.stderr)
            return 1

        # Special case
        if (
            EQUAL(drv.GetName(), "ESRI Shapefile")
            and output_lyr_name is None
            and EQUAL(os.path.splitext(output_ds_name)[1], ".SHP")
        ):
            output_lyr_name = os.path.splitext(os.path.basename(output_ds_name))[0]

        if output_lyr_name is None:
            print("-output_lyr should be specified", file=sys.stderr)
            return 1

        output_lyr = CreateLayer(
            output_ds,
            output_lyr_name,
            srs,
            geom_type,
            lco,
            input_lyr,
            input_fields,
            method_lyr,
            method_fields,
            opt,
        )
        if output_lyr is None:
            return 1
    else:
        drv = output_ds.GetDriver()

        if output_lyr_name is None:
            cnt = output_ds.GetLayerCount()
            if cnt != 1:
                print(
                    "Result datasource has not a single layer, so you should specify its name with -output_lyr",
                    file=sys.stderr,
                )
                return 1
            output_lyr = output_ds.GetLayer(0)
            output_lyr_name = output_lyr.GetName()
        else:
            output_lyr = output_ds.GetLayerByName(output_lyr_name)

            if output_lyr is None:
                if (
                    EQUAL(drv.GetName(), "ESRI Shapefile")
                    and EQUAL(os.path.splitext(output_ds_name)[1], ".SHP")
                    and not EQUAL(
                        output_lyr_name,
                        os.path.splitext(os.path.basename(output_ds_name))[0],
                    )
                ):
                    print(
                        'Cannot create layer "%s" in a shapefile called "%s"'
                        % (output_lyr_name, output_ds_name),
                        file=sys.stderr,
                    )
                    return 1

                output_lyr = CreateLayer(
                    output_ds,
                    output_lyr_name,
                    srs,
                    geom_type,
                    lco,
                    input_lyr,
                    input_fields,
                    method_lyr,
                    method_fields,
                    opt,
                )
                if output_lyr is None:
                    return 1

        if overwrite:
            cnt = output_ds.GetLayerCount()
            iLayer = None  # initialize in case there are no loop iterations
            for iLayer in range(cnt):
                poLayer = output_ds.GetLayer(iLayer)
                if poLayer is not None and poLayer.GetName() == output_lyr_name:
                    break
            if iLayer != cnt:
                if output_ds.DeleteLayer(iLayer) != 0:
                    print(
                        "DeleteLayer() failed when overwrite requested.",
                        file=sys.stderr,
                    )
                    return 1

                output_lyr = CreateLayer(
                    output_ds,
                    output_lyr_name,
                    srs,
                    geom_type,
                    lco,
                    input_lyr,
                    input_fields,
                    method_lyr,
                    method_fields,
                    opt,
                )
                if output_lyr is None:
                    return 1

    op = getattr(input_lyr, op_str)
    if not quiet:
        ret = op(method_lyr, output_lyr, options=opt, callback=gdal.TermProgress_nocb)
    else:
        ret = op(method_lyr, output_lyr, options=opt)

    input_ds = None
    method_ds = None
    output_ds = None

    if ret != 0:
        print("An error occurred during %s operation" % op_str, file=sys.stderr)
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
###############################################################################
#
# Project:  GDAL/OGR
# Purpose:  Test compliance of GeoParquet file
# Author:   Even Rouault <even.rouault at spatialys.com>
#
###############################################################################
# Copyright (c) 2023, Even Rouault <even.rouault at spatialys.com>
#
# SPDX-License-Identifier: MIT
###############################################################################

import json
import os
import sys

from osgeo import gdal, ogr, osr

geoparquet_schemas = {}

map_ogr_geom_type_to_geoparquet = {
    ogr.wkbPoint: "Point",
    ogr.wkbLineString: "LineString",
    ogr.wkbPolygon: "Polygon",
    ogr.wkbMultiPoint: "MultiPoint",
    ogr.wkbMultiLineString: "MultiLineString",
    ogr.wkbMultiPolygon: "MultiPolygon",
    ogr.wkbGeometryCollection: "GeometryCollection",
    ogr.wkbPoint25D: "Point",
    ogr.wkbLineString25D: "LineString Z",
    ogr.wkbPolygon25D: "Polygon Z",
    ogr.wkbMultiPoint25D: "MultiPoint Z",
    ogr.wkbMultiLineString25D: "MultiLineString Z",
    ogr.wkbMultiPolygon25D: "MultiPolygon Z",
    ogr.wkbGeometryCollection25D: "GeometryCollection Z",
}

map_remote_resources = {}


def _try_local_proj_schema(uri):
    """Try to resolve a proj.org schema URI from local PROJ data directory."""
    prefix = "https://proj.org/schemas/"
    if not uri.startswith(prefix):
        return None
    relative = uri[len(prefix) :]
    try:
        for search_path in osr.GetPROJSearchPaths():
            local_path = os.path.join(search_path, os.path.basename(relative))
            if os.path.isfile(local_path):
                with open(local_path, "rb") as f:
                    return f.read()
    except OSError:
        pass
    return None


class GeoParquetValidator:
    def __init__(self, filename, check_data=False, local_schema=None):
        self.filename = filename
        self.check_data = check_data
        self.local_schema = local_schema
        self.ds = None
        self.errors = []

    def check(self):
        with gdal.ExceptionMgr(useExceptions=True), ogr.ExceptionMgr(
            useExceptions=True
        ), osr.ExceptionMgr(useExceptions=True):
            self._check()
            self.ds = None

    def _error(self, msg):
        if len(self.errors) == 100:
            self.errors.append(
                f"{self.filename}: too many errors. No longer emitting ones"
            )
        elif len(self.errors) < 100:
            self.errors.append(f"{self.filename}: {msg}")

    def _check_counterclockwise(self, g, row):
        gt = ogr.GT_Flatten(g.GetGeometryType())
        if gt == ogr.wkbPolygon:
            for idx in range(g.GetGeometryCount()):
                ring = g.GetGeometryRef(idx)
                if idx == 0 and ring.IsClockwise():
                    self._error(
                        f"Exterior ring of geometry at row {row} has invalid orientation"
                    )
                elif idx > 0 and not ring.IsClockwise():
                    self._error(
                        f"Interior ring of geometry at row {row} has invalid orientation"
                    )
        elif gt in (ogr.wkbMultiPolygon, ogr.wkbGeometryCollection):
            for idx in range(g.GetGeometryCount()):
                subgeom = g.GetGeometryRef(idx)
                self._check_counterclockwise(subgeom, row)

    def _validate(self, schema, instance):
        from importlib.metadata import version

        import jsonschema

        jsonschema_version = version("jsonschema")

        def versiontuple(v):
            return tuple(map(int, (v.split("."))))

        # jsonschema 4.18 deprecates automatic resolution of "$ref" for security
        # reason. Use a custom retrieve method.
        # Cf https://python-jsonschema.readthedocs.io/en/latest/referencing/#automatically-retrieving-resources-over-http
        if versiontuple(jsonschema_version) >= (4, 18):
            from referencing import Registry, Resource

            def retrieve_remote_file(uri: str):
                if not uri.startswith("http://") and not uri.startswith("https://"):
                    raise Exception(f"Cannot retrieve {uri}")

                if uri not in map_remote_resources:
                    local_content = _try_local_proj_schema(uri)
                    if local_content is not None:
                        map_remote_resources[uri] = local_content
                    else:
                        import urllib.request

                        response = urllib.request.urlopen(
                            urllib.request.Request(uri, headers={"User-Agent": "GDAL"})
                        ).read()
                        map_remote_resources[uri] = response

                return Resource.from_contents(json.loads(map_remote_resources[uri]))

            registry = Registry(retrieve=retrieve_remote_file)
            validator_cls = jsonschema.validators.validator_for(schema)
            validator_cls(schema, registry=registry).validate(instance)

        else:
            # jsonschema < 4.18
            jsonschema.validate(instance=instance, schema=schema)

    def _check(self):

        if gdal.GetDriverByName("Parquet") is None:
            return self._error("Parquet driver not available")

        try:
            import jsonschema  # noqa: F401
        except ImportError:
            return self._error(
                "jsonschema Python module not available. Try 'pip install jsonschema'"
            )

        try:
            self.ds = gdal.OpenEx(self.filename, allowed_drivers=["Parquet"])
        except Exception as e:
            return self._error("is not a Parquet file: %s" % str(e))

        lyr = self.ds.GetLayer(0)
        geo = lyr.GetMetadataItem("geo", "_PARQUET_METADATA_")
        if geo is None:
            return self._error("'geo' file metadata item missing")

        try:
            geo_j = json.loads(geo)
        except Exception as e:
            return self._error(
                "'geo' metadata item is not valid JSON. Value of 'geo' = '%s'. Exception = '%s' "
                % (geo, str(e))
            )

        if "version" not in geo_j:
            self._error(
                "'geo' metadata item lacks a 'version' member. Value of 'geo' = '%s'"
                % geo
            )
            if not self.local_schema:
                return

            version = None
        else:
            version = geo_j["version"]
            if not isinstance(version, str):
                return self._error(
                    "'geo[\"version\"]' is not a string. Value of 'geo' = '%s'" % geo
                )

        if self.local_schema:
            schema_j = json.loads(open(self.local_schema, "rb").read())
        else:
            schema_url = f"https://github.com/opengeospatial/geoparquet/releases/download/v{version}/schema.json"

            if schema_url not in geoparquet_schemas:
                import urllib.request

                try:
                    response = urllib.request.urlopen(
                        urllib.request.Request(
                            schema_url, headers={"User-Agent": "GDAL"}
                        )
                    ).read()
                except Exception as e:
                    return self._error(
                        f"Cannot download GeoParquet JSON schema from {schema_url}. Exception = {repr(e)}"
                    )

                try:
                    geoparquet_schemas[schema_url] = json.loads(response)
                except Exception as e:
                    return self._error(
                        f"Failed to read GeoParquet schema at {schema_url} as JSON. Schema content = '{response}'. Exception = {repr(e)}"
                    )

            schema_j = geoparquet_schemas[schema_url]

        try:
            self._validate(schema_j, geo_j)
        except Exception as e:
            self._error(
                "'geo' metadata item fails to validate its schema: %s'" % str(e)
            )

        if "primary_column" not in geo_j:
            return self._error('geo["primary_column"] missing')
        primary_column = geo_j["primary_column"]

        if "columns" not in geo_j:
            return self._error('geo["columns"] missing')
        columns = geo_j["columns"]

        if primary_column not in columns:
            self._error(
                f'geo["primary_column"] = {primary_column} is not in listed in geo["columns"'
            )

        ogr_geom_field_names = [
            lyr.GetLayerDefn().GetGeomFieldDefn(i).GetName()
            for i in range(lyr.GetLayerDefn().GetGeomFieldCount())
        ]

        for column_name, column_def in columns.items():
            if column_name not in ogr_geom_field_names:
                self._error(
                    f'geo["columns"] lists a {column_name} column which is not found in the Parquet fields'
                )

            self._check_column_metadata(column_name, column_def)

        if self.check_data:
            self._check_data(lyr, columns)

    def _check_column_metadata(self, column_name, column_def):
        srs = None
        if "crs" in column_def:
            crs = column_def["crs"]
            if crs and (
                osr.GetPROJVersionMajor() * 100 + osr.GetPROJVersionMinor() >= 602
            ):
                srs = osr.SpatialReference()
                try:
                    if isinstance(crs, dict):
                        srs.SetFromUserInput(json.dumps(crs))
                    elif isinstance(crs, str):
                        srs.SetFromUserInput(crs)
                except Exception as e:
                    self._error(f"crs {crs} cannot be parsed: %s" % str(e))
        else:
            srs = osr.SpatialReference()
            srs.ImportFromEPSG(4326)

        if "bbox" in column_def:
            bbox = column_def["bbox"]
            if bbox:
                if len(bbox) == 4:
                    minx_idx = 0
                    miny_idx = 1
                    maxx_idx = 2
                    maxy_idx = 3

                elif len(bbox) == 6:
                    minx_idx = 0
                    miny_idx = 1
                    minz_idx = 2
                    maxx_idx = 3
                    maxy_idx = 4
                    maxz_idx = 5

                    if bbox[maxz_idx] < bbox[minz_idx]:
                        self._error(f"bbox[{maxz_idx}] < bbox[{minz_idx}]")

                else:
                    assert False, "Unexpected len(bbox)"

                if srs and srs.IsGeographic():
                    if abs(bbox[minx_idx]) > 180:
                        self._error(f"abs(bbox[{minx_idx}]) > 180")
                    if abs(bbox[miny_idx]) > 90:
                        self._error(f"abs(bbox[{miny_idx}]) > 90")
                    if abs(bbox[maxx_idx]) > 180:
                        self._error(f"abs(bbox[{maxx_idx}]) > 180")
                    if abs(bbox[maxy_idx]) > 90:
                        self._error(f"abs(bbox[{maxy_idx}]) > 90")
                if srs and srs.IsProjected():
                    if bbox[maxx_idx] < bbox[minx_idx]:
                        self._error(f"bbox[{maxx_idx}] < bbox[{minx_idx}]")

                if bbox[maxy_idx] < bbox[miny_idx]:
                    self._error(f"bbox[{maxy_idx}] < bbox[{miny_idx}]")

            if "geometry_types" in column_def:
                geometry_types = column_def["geometry_types"]
                set_geometry_types = set()
                for geometry_type in geometry_types:
                    if geometry_type in set_geometry_types:
                        self._error(
                            f"{geometry_type} is declared several times in geometry_types[]"
                        )

    def _check_data_with_high_level_ogr_api(self, lyr, columns):
        """Use for GeoArrow validation"""

        if gdal.VersionInfo() < "3090000":
            raise Exception("GDAL 3.9 or later required for GeoArrow data validation")

        lyr.SetIgnoredFields(
            [
                lyr.GetLayerDefn().GetFieldDefn(i).GetName()
                for i in range(lyr.GetLayerDefn().GetFieldCount())
            ]
        )

        list_of_set_geometry_types = []
        orientations = []
        encodings = []
        assert len(columns) == lyr.GetLayerDefn().GetGeomFieldCount()
        for i in range(lyr.GetLayerDefn().GetGeomFieldCount()):
            column_def = columns[lyr.GetLayerDefn().GetGeomFieldDefn(i).GetName()]

            encodings.append(column_def["encoding"])

            if "geometry_types" in column_def:
                geometry_types = column_def["geometry_types"]
                set_geometry_types = set(geometry_types)
            else:
                set_geometry_types = set()
            list_of_set_geometry_types.append(set_geometry_types)

            if "orientation" in column_def:
                orientation = column_def["orientation"]
            else:
                orientation = None
            orientations.append(orientation)

        map_ogr_geom_type_to_geoarrow_encoding = {
            ogr.wkbPoint: "point",
            ogr.wkbLineString: "linestring",
            ogr.wkbPolygon: "polygon",
            ogr.wkbMultiPoint: "multipoint",
            ogr.wkbMultiLineString: "multilinestring",
            ogr.wkbMultiPolygon: "multipolygon",
        }

        for row, f in enumerate(lyr):
            for i in range(lyr.GetLayerDefn().GetGeomFieldCount()):
                g = f.GetGeomFieldRef(i)
                if g:
                    ogr_geom_type = g.GetGeometryType()
                    if ogr_geom_type not in map_ogr_geom_type_to_geoparquet:
                        self._error(
                            f"Geometry at row {row} is of unexpected type for GeoParquet: %s"
                            % g.GetGeometryName()
                        )
                    elif list_of_set_geometry_types[i]:
                        geoparquet_geom_type = map_ogr_geom_type_to_geoparquet[
                            ogr_geom_type
                        ]
                        if geoparquet_geom_type not in list_of_set_geometry_types[i]:
                            self._error(
                                f"Geometry at row {row} is of type {geoparquet_geom_type}, but not listed in geometry_types[]"
                            )

                    ogr_flat_geom_type = ogr.GT_Flatten(ogr_geom_type)
                    if ogr_flat_geom_type not in map_ogr_geom_type_to_geoarrow_encoding:
                        self._error(
                            f"Geometry at row {row} is of unexpected type for a GeoArrow encoding of GeoParquet: %s"
                            % g.GetGeometryName()
                        )
                    elif (
                        map_ogr_geom_type_to_geoarrow_encoding[ogr_flat_geom_type]
                        != encodings[i]
                    ):
                        self._error(
                            f"Geometry at row {row} is a %s but does not match the declared encoding of %s"
                            % (ogr.GeometryTypeToName(ogr_geom_type), encodings[i])
                        )

                    if orientations[i] == "counterclockwise":
                        self._check_counterclockwise(g, row)

    def _check_data(self, lyr, columns):

        # For non-WKB encoding, just use the high level OGR API
        use_high_level_ogr_api = False
        for _, column_def in columns.items():
            if column_def["encoding"] != "WKB":
                use_high_level_ogr_api = True
                break

        if use_high_level_ogr_api:
            self._check_data_with_high_level_ogr_api(lyr, columns)
            return

        lyr.SetIgnoredFields(
            [
                lyr.GetLayerDefn().GetFieldDefn(i).GetName()
                for i in range(lyr.GetLayerDefn().GetFieldCount())
            ]
        )

        stream = lyr.GetArrowStreamAsNumPy(options=["USE_MASKED_ARRAYS=NO"])
        row = 0
        for batch in stream:
            row_backup = row
            rows_in_batch = 0
            for column_name, column_def in columns.items():
                if "geometry_types" in column_def:
                    geometry_types = column_def["geometry_types"]
                    set_geometry_types = set(geometry_types)
                else:
                    set_geometry_types = set()

                if "orientation" in column_def:
                    orientation = column_def["orientation"]
                else:
                    orientation = None

                row = row_backup
                rows_in_batch = len(batch[column_name])
                for geom in batch[column_name]:
                    g = None
                    try:
                        if geom is not None:
                            g = ogr.CreateGeometryFromWkb(geom)
                    except Exception as e:
                        self._error(
                            f"Invalid WKB geometry at row {row} for column {column_name}: %s"
                            % str(e)
                        )

                    if g:
                        ogr_geom_type = g.GetGeometryType()
                        if ogr_geom_type not in map_ogr_geom_type_to_geoparquet:
                            self._error(
                                f"Geometry at row {row} is of unexpected type for GeoParquet: %s"
                                % g.GetGeometryName()
                            )
                        elif set_geometry_types:
                            geoparquet_geom_type = map_ogr_geom_type_to_geoparquet[
                                ogr_geom_type
                            ]
                            if geoparquet_geom_type not in set_geometry_types:
                                self._error(
                                    f"Geometry at row {row} is of type {geoparquet_geom_type}, but not listed in geometry_types[]"
                                )

                        # ogr.Geometry.IsClockwise() added in GDAL 3.8
                        if (
                            hasattr(g, "IsClockwise")
                            and orientation == "counterclockwise"
                        ):
                            self._check_counterclockwise(g, row)

                    row += 1

            row += rows_in_batch


def check(filename, check_data=False, local_schema=None):
    """Validate a file against GeoParquet specification

    Parameters
    ----------
    filename:
        File to validate
    check_data:
        Set to True to check geometry content in addition to metadata.
    local_schema:
        Path to local schema (if not specified, it will be retrieved at https://github.com/opengeospatial/geoparquet)

    Returns
    -------
    A list of error messages, or an empty list if no error
    """
    checker = GeoParquetValidator(
        filename, check_data=check_data, local_schema=local_schema
    )
    checker.check()
    return checker.errors


def Usage():
    print(
        "Usage: validate_geoparquet.py [--check-data] [--schema FILENAME] my_geo.parquet"
    )
    print("")
    print("--check-data: validate data in addition to metadata")
    print(
        "--schema: path to GeoParquet JSON schema. If not specified, retrieved from the network"
    )
    return 2


def main(argv=sys.argv):
    filename = None
    if len(argv) == 1:
        return Usage()
    check_data = False
    local_schema = None
    i = 1
    while i < len(argv):
        arg = argv[i]
        if arg == "--check-data":
            check_data = True
        elif arg == "--schema":
            local_schema = argv[i + 1]
            i += 1
        elif arg[0] == "-":
            return Usage()
        else:
            filename = arg

        i += 1

    if filename is None:
        return Usage()
    errors = check(filename, check_data=check_data, local_schema=local_schema)
    if errors:
        for msg in errors:
            print(msg)
        return 1
    return 0


if __name__ == "__main__":
    gdal.UseExceptions()
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
###############################################################################
#
# Project:  OGR Python samples
# Purpose:  Produce a graticule (grid) dataset.
# Author:   Frank Warmerdam, warmerdam@pobox.com
#
###############################################################################
# Copyright (c) 2003, Frank Warmerdam <warmerdam@pobox.com>
#
# SPDX-License-Identifier: MIT
###############################################################################

import sys

from osgeo import ogr, osr


#############################################################################
def float_range(*args):
    start = 0.0
    step = 1.0
    if len(args) == 1:
        (stop,) = args
    elif len(args) == 2:
        start, stop = args
    elif len(args) == 3:
        start, stop, step = args
    else:
        raise TypeError("float_range needs 1-3 float arguments")

    the_range = []
    steps = (stop - start) / step
    if steps != int(steps):
        steps = steps + 1.0
    for i in range(int(steps)):
        the_range.append(i * step + start)

    return the_range


#############################################################################
def Usage():
    print("Usage: mkgraticule [-connected] [-s stepsize] [-substep substepsize]")
    print("         [-t_srs srs] [-range xmin ymin xmax ymax] outfile")
    print("")
    print(
        "Defaults to: not connected, step & substep of 5.0, no SRS, range -180,-90 180,90"
    )
    return 2


def main(argv=sys.argv):
    # Argument processing.
    t_srs = None
    stepsize = 5.0
    substepsize = 5.0
    connected = 0
    outfile = None

    xmin = -180
    xmax = 180
    ymin = -90
    ymax = 90

    i = 1
    while i < len(argv):
        if argv[i] == "-connected":
            connected = 1
        elif argv[i] == "-t_srs":
            i = i + 1
            t_srs = argv[i]
        elif argv[i] == "-s":
            i = i + 1
            stepsize = float(argv[i])
        elif argv[i] == "-substep":
            i = i + 1
            substepsize = float(argv[i])
        elif argv[i] == "-range":
            xmin = float(argv[i + 1])
            ymin = float(argv[i + 2])
            xmax = float(argv[i + 3])
            ymax = float(argv[i + 4])
            i = i + 4
        elif argv[i][0] == "-":
            return Usage()
        elif outfile is None:
            outfile = argv[i]
        else:
            return Usage()

        i = i + 1

    if outfile is None:
        print("""\nNo outfile specified, e.g. 'graticule.shp'.\n""")
        return Usage()

    if substepsize > stepsize:
        substepsize = stepsize

    # -
    # Do we have an alternate SRS?

    ct = None

    if t_srs is not None:
        t_srs_o = osr.SpatialReference()
        t_srs_o.SetFromUserInput(t_srs)

        s_srs_o = osr.SpatialReference()
        s_srs_o.SetFromUserInput("WGS84")

        ct = osr.CoordinateTransformation(s_srs_o, t_srs_o)
    else:
        t_srs_o = osr.SpatialReference()
        t_srs_o.SetFromUserInput("WGS84")

    # -
    # Create graticule file.

    drv = ogr.GetDriverByName("ESRI Shapefile")

    try:
        drv.DeleteDataSource(outfile)
    except Exception:
        pass

    ds = drv.CreateDataSource(outfile)
    layer = ds.CreateLayer("out", geom_type=ogr.wkbLineString, srs=t_srs_o)

    #########################################################################
    # Not connected case.  Produce individual segments are these are going to
    # be more resilient in the face of reprojection errors.

    if not connected:
        #########################################################################
        # Generate lines of latitude.

        feat = ogr.Feature(feature_def=layer.GetLayerDefn())
        geom = ogr.Geometry(type=ogr.wkbLineString)

        for lat in float_range(ymin, ymax + stepsize / 2, stepsize):
            for long_ in float_range(xmin, xmax - substepsize / 2, substepsize):

                geom.SetPoint(0, long_, lat)
                geom.SetPoint(1, long_ + substepsize, lat)

                err = 0
                if ct is not None:
                    err = geom.Transform(ct)

                if err == 0:
                    feat.SetGeometry(geom)
                    layer.CreateFeature(feat)

        #########################################################################
        # Generate lines of longitude

        for long_ in float_range(xmin, xmax + stepsize / 2, stepsize):
            for lat in float_range(ymin, ymax - substepsize / 2, substepsize):
                geom.SetPoint(0, long_, lat)
                geom.SetPoint(1, long_, lat + substepsize)

                err = 0
                if ct is not None:
                    err = geom.Transform(ct)

                if err == 0:
                    feat.SetGeometry(geom)
                    layer.CreateFeature(feat)

    #########################################################################
    # Connected case - produce one polyline for each complete line of latitude
    # or longitude.

    if connected:
        #########################################################################
        # Generate lines of latitude.

        feat = ogr.Feature(feature_def=layer.GetLayerDefn())

        for lat in float_range(ymin, ymax + stepsize / 2, stepsize):

            geom = ogr.Geometry(type=ogr.wkbLineString)

            for long_ in float_range(xmin, xmax + substepsize / 2, substepsize):
                geom.AddPoint(long_, lat)

            err = 0
            if ct is not None:
                err = geom.Transform(ct)

            if err == 0:
                feat.SetGeometry(geom)
                layer.CreateFeature(feat)

        #########################################################################
        # Generate lines of longitude

        for long_ in float_range(xmin, xmax + stepsize / 2, stepsize):

            geom = ogr.Geometry(type=ogr.wkbLineString)

            for lat in float_range(ymin, ymax + substepsize / 2, substepsize):
                geom.AddPoint(long_, lat)

            err = 0
            if ct is not None:
                err = geom.Transform(ct)

            if err == 0:
                feat.SetGeometry(geom)
                layer.CreateFeature(feat)

    #############################################################################
    # Cleanup

    feat = None
    geom = None

    ds.Destroy()
    ds = None

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
###############################################################################
#
# Project:  GDAL/OGR samples
# Purpose:  Extract a JPEG file from a JPEG-in-TIFF tile/strip
# Author:   Even Rouault <even dot rouault at spatialys.com>
#
###############################################################################
# Copyright (c) 2014, Even Rouault <even dot rouault at spatialys.com>
#
# SPDX-License-Identifier: MIT
###############################################################################

import sys

from osgeo import gdal


def Usage():
    print("Usage: jpeg_in_tiff_extract.py in.tif out.jpg [tile_x tile_y [band_nbr]]")
    print("")
    print("Extract a JPEG file from a JPEG-in-TIFF tile/strip.")
    print("If tile_x tile_y are not specified, then all tiles/strips are extracted")
    print("in filenames out_[bandnbr_]tx_ty.jpg")
    print("")
    return 2


###############################################################


def extract_tile(ds, src_band_nbr, tile_x, tile_y, jpg_filename):

    block_offset = ds.GetRasterBand(src_band_nbr).GetMetadataItem(
        "BLOCK_OFFSET_%d_%d" % (tile_x, tile_y), "TIFF"
    )
    block_size = ds.GetRasterBand(src_band_nbr).GetMetadataItem(
        "BLOCK_SIZE_%d_%d" % (tile_x, tile_y), "TIFF"
    )
    if block_offset is None or block_size is None:
        print("ERROR: Cannot find block (%d,%d)" % (tile_x, tile_y))
        return 1

    jpegtables = ds.GetRasterBand(src_band_nbr).GetMetadataItem("JPEGTABLES", "TIFF")
    if jpegtables is not None:
        if (
            (len(jpegtables) % 2) != 0
            or jpegtables[0:4] != "FFD8"
            or jpegtables[-2:] != "D9"
        ):
            print("ERROR: Invalid JPEG tables")
            print(jpegtables)
            return 1

        # Remove final D9
        jpegtables = jpegtables[0:-2]

    tiff_f = gdal.VSIFOpenL(ds.GetDescription(), "rb")
    if tiff_f is None:
        print("ERROR: Cannot reopen %s" % ds.GetDescription())
        return 1

    out_f = gdal.VSIFOpenL(jpg_filename, "wb")
    if out_f is None:
        print("ERROR: Cannot create %s" % jpg_filename)
        gdal.VSIFCloseL(tiff_f)
        return 1

    # Write JPEG tables
    if jpegtables is not None:
        for i in range(int(len(jpegtables) / 2)):
            c1 = ord(jpegtables[2 * i])
            c2 = ord(jpegtables[2 * i + 1])
            if c1 >= ord("0") and c1 <= ord("9"):
                val = c1 - ord("0")
            else:
                val = (c1 - ord("A")) + 10
            val = val * 16
            if c2 >= ord("0") and c2 <= ord("9"):
                val = val + (c2 - ord("0"))
            else:
                val = val + (c2 - ord("A")) + 10
            gdal.VSIFWriteL(chr(val), 1, 1, out_f)
    else:
        gdal.VSIFWriteL(chr(0xFF), 1, 1, out_f)
        gdal.VSIFWriteL(chr(0xD8), 1, 1, out_f)

    # Write Adobe APP14 marker if necessary
    interleave = ds.GetMetadataItem("INTERLEAVE", "IMAGE_STRUCTURE")
    photometric = ds.GetMetadataItem("COMPRESSION", "IMAGE_STRUCTURE")
    if interleave == "PIXEL" and photometric == "JPEG" and ds.RasterCount == 3:
        adobe_app14 = [
            0xFF,
            0xEE,
            0x00,
            0x0E,
            0x41,
            0x64,
            0x6F,
            0x62,
            0x65,
            0x00,
            0x64,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
        ]
        for c in adobe_app14:
            gdal.VSIFWriteL(chr(c), 1, 1, out_f)

    # Write JPEG codestream
    # skip leading 0xFF 0xD8
    gdal.VSIFSeekL(tiff_f, int(block_offset) + 2, 0)
    data = gdal.VSIFReadL(1, int(block_size) - 2, tiff_f)
    gdal.VSIFCloseL(tiff_f)
    gdal.VSIFWriteL(data, 1, len(data), out_f)

    gdal.VSIFCloseL(out_f)

    aux_xml_filename = "%s.aux.xml" % jpg_filename
    gt = ds.GetGeoTransform()
    srs = ds.GetProjectionRef()
    if srs is not None and srs != "":
        sub_gt = [gt[i] for i in range(6)]
        blockxsize, blockysize = ds.GetRasterBand(1).GetBlockSize()
        sub_gt[0] = gt[0] + tile_x * blockxsize * gt[1]
        sub_gt[3] = gt[3] + tile_y * blockysize * gt[5]

        out_f = gdal.VSIFOpenL(aux_xml_filename, "wb")
        if out_f is None:
            print("ERROR: Cannot create %s" % aux_xml_filename)
            return 1
        content = """<PAMDataset>
    <SRS>%s</SRS>
    <GeoTransform>%.18g,%.18g,%.18g,%.18g,%.18g,%.18g</GeoTransform>
    </PAMDataset>
    """ % (
            srs,
            sub_gt[0],
            sub_gt[1],
            sub_gt[2],
            sub_gt[3],
            sub_gt[4],
            sub_gt[5],
        )
        gdal.VSIFWriteL(content, 1, len(content), out_f)
        gdal.VSIFCloseL(out_f)
    else:
        gdal.Unlink("%s.aux.xml" % jpg_filename)

    return 0


###############################################################


def jpeg_in_tiff_extract(argv):

    if len(argv) < 2:
        print("ERROR: Not enough arguments")
        return Usage()

    tiff_filename = argv[0]
    jpg_filename = argv[1]
    if len(argv) >= 3:
        tile_x = int(argv[2])
        tile_y = int(argv[3])
        if len(argv) == 5:
            band_nbr = int(argv[4])
        else:
            band_nbr = None
    else:
        tile_x = None
        tile_y = None

    radix_jpg_filename = jpg_filename
    extensions = [".jpg", ".jpeg", ".JPG", ".JPEG"]
    extension = None
    for ext in extensions:
        pos = radix_jpg_filename.find(ext)
        if pos >= 0:
            extension = ext
            radix_jpg_filename = radix_jpg_filename[0:pos]
            break
    if pos < 0:
        print("ERROR: %s should end with .jpg/.jpeg" % jpg_filename)
        return 1

    ds = gdal.Open(tiff_filename)
    if ds is None:
        print("ERROR: Cannot open %s" % tiff_filename)
        return 1

    if ds.GetDriver() is None or ds.GetDriver().GetDescription() != "GTiff":
        print("ERROR: %s is not a TIFF dataset." % tiff_filename)
        return 1

    photometric = ds.GetMetadataItem("COMPRESSION", "IMAGE_STRUCTURE")
    interleave = ds.GetMetadataItem("INTERLEAVE", "IMAGE_STRUCTURE")

    if photometric != "JPEG" and photometric != "YCbCr JPEG":
        print("ERROR: %s is not a JPEG-compressed TIFF dataset." % tiff_filename)
        return 1

    blockxsize, blockysize = ds.GetRasterBand(1).GetBlockSize()
    if blockysize == 1:
        blockysize = ds.RasterYSize
    block_in_row = (ds.RasterXSize + blockxsize - 1) / blockxsize
    block_in_col = (ds.RasterYSize + blockysize - 1) / blockysize

    # Extract single tile ?
    if tile_x is not None:

        if tile_x < 0 or tile_x >= block_in_row:
            print(
                "ERROR: Invalid tile_x : %d. Should be >= 0 and < %d"
                % (tile_x, block_in_row)
            )
            return 1
        if tile_y < 0 or tile_y >= block_in_col:
            print(
                "ERROR: Invalid tile_y : %d. Should be >= 0 and < %d"
                % (tile_y, block_in_col)
            )
            return 1

        if ds.RasterCount > 1:
            if interleave == "PIXEL":
                if band_nbr is not None:
                    print(
                        "ERROR: For a INTERLEAVE=PIXEL dataset, band_nbr should NOT be specified"
                    )
                    return 1
            else:
                if band_nbr is None:
                    print(
                        "ERROR: For a INTERLEAVE=BAND dataset, band_nbr should be specified"
                    )
                    return 1

        if band_nbr is not None:
            if band_nbr < 1 or band_nbr >= ds.RasterCount:
                print(
                    "ERROR: Invalid band_nbr : %d. Should be >= 1 and <= %d"
                    % (tile_y, ds.RasterCount)
                )
                return 1

        if band_nbr is not None:
            src_band_nbr = band_nbr
        else:
            src_band_nbr = 1

        return extract_tile(ds, src_band_nbr, tile_x, tile_y, jpg_filename)

    # Extract all tiles
    else:
        if ds.RasterCount == 1 or interleave == "PIXEL":
            for tile_y in range(block_in_col):
                for tile_x in range(block_in_row):
                    filename = "%s_%d_%d%s" % (
                        radix_jpg_filename,
                        tile_x,
                        tile_y,
                        extension,
                    )
                    ret = extract_tile(ds, 1, tile_x, tile_y, filename)
                    if ret != 0:
                        return ret
        else:
            for src_band_nbr in range(ds.RasterCount):
                for tile_y in range(block_in_col):
                    for tile_x in range(block_in_row):
                        filename = "%s_%d_%d_%d%s" % (
                            radix_jpg_filename,
                            src_band_nbr + 1,
                            tile_x,
                            tile_y,
                            extension,
                        )
                        ret = extract_tile(
                            ds, src_band_nbr + 1, tile_x, tile_y, filename
                        )
                        if ret != 0:
                            return ret
        return 0


def main(argv=sys.argv):
    gdal.GeneralCmdLineProcessor(argv)
    return jpeg_in_tiff_extract(argv[1:])


if __name__ == "__main__":
    sys.exit(main(sys.argv))
